#include <security/pam_modules.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <libzfs.h>
#include <errno.h>

// Parse PAM arguments (e.g., "compression=lz4,quota=10G")
static void parse_properties(const char *input, nvlist_t **props) {
    char *token, *key, *value;
    char *input_copy = strdup(input);
    char *rest = input_copy;

    *props = NULL;
    if (nvlist_alloc(props, NV_UNIQUE_NAME, 0) != 0) {
        free(input_copy);
        return;
    }

    while ((token = strtok_r(rest, ",", &rest))) {
        key = token;
        value = strchr(token, '=');
        if (!value) continue;
        *value++ = '\0';

        if (nvlist_add_string(*props, key, value) != 0) {
            pam_syslog(NULL, LOG_ERR, "Failed to add property %s=%s", key, value);
        }
    }
    free(input_copy);
}

PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags, int argc, const char **argv) {
    const char *user;
    int retval;
    libzfs_handle_t *zfs_handle;
    zfs_handle_t *home_dataset;
    char dataset_name[256];
    char home_dir[256];
    nvlist_t *props = NULL;

    // Get the username
    if ((retval = pam_get_user(pamh, &user, NULL)) != PAM_SUCCESS) {
        pam_syslog(pamh, LOG_ERR, "Failed to get username");
        return retval;
    }

    // Parse module arguments (e.g., "compression=zstd,quota=20G")
    for (int i = 0; i < argc; i++) {
        if (strncmp(argv[i], "properties=", 11) == 0) {
            parse_properties(argv[i] + 11, &props);
        }
    }

    // Initialize libzfs
    if (!(zfs_handle = libzfs_init())) {
        pam_syslog(pamh, LOG_ERR, "Failed to initialize ZFS library");
        return PAM_SESSION_ERR;
    }

    // Construct dataset name (e.g., "rpool/home/%s")
    snprintf(dataset_name, sizeof(dataset_name), "rpool/home/%s", user);

    // Skip if dataset already exists
    if (zfs_dataset_exists(zfs_handle, dataset_name, ZFS_TYPE_FILESYSTEM)) {
        libzfs_fini(zfs_handle);
        return PAM_SUCCESS;
    }

    // Create the dataset with custom properties
    if (zfs_create_annotated(zfs_handle, dataset_name, ZFS_TYPE_FILESYSTEM, props) != 0) {
        pam_syslog(pamh, LOG_ERR, "Failed to create ZFS dataset %s: %s",
                  dataset_name, libzfs_error_description(zfs_handle));
        if (props) nvlist_free(props);
        libzfs_fini(zfs_handle);
        return PAM_SESSION_ERR;
    }

    // Open dataset to enforce mountpoint
    home_dataset = zfs_open(zfs_handle, dataset_name, ZFS_TYPE_FILESYSTEM);
    if (!home_dataset) {
        pam_syslog(pamh, LOG_ERR, "Failed to open dataset %s: %s",
                  dataset_name, libzfs_error_description(zfs_handle));
        if (props) nvlist_free(props);
        libzfs_fini(zfs_handle);
        return PAM_SESSION_ERR;
    }

    // Ensure mountpoint is /home/%s (override properties if needed)
    snprintf(home_dir, sizeof(home_dir), "/home/%s", user);
    if (zfs_prop_set(home_dataset, "mountpoint", home_dir) != 0) {
        pam_syslog(pamh, LOG_ERR, "Failed to set mountpoint for %s: %s",
                  dataset_name, libzfs_error_description(zfs_handle));
    }

    // Cleanup
    zfs_close(home_dataset);
    if (props) nvlist_free(props);
    libzfs_fini(zfs_handle);

    // Fallback: Ensure directory exists (ZFS should handle this)
    mkdir(home_dir, 0700);
    chown(home_dir, getpwnam(user)->pw_uid, getpwnam(user)->pw_gid);

    return PAM_SUCCESS;
}
